@@ -33,11 +33,13 @@
 #include "terminal-widget.h"
 #include "terminal-profile.h"
 #include "terminal.h"
+#include "skey-popup.h"
 #include <libgnome/gnome-util.h> /* gnome_util_user_shell */
 #include <libgnome/gnome-url.h> /* gnome_url_show */
 #include <string.h>
 #include <stdlib.h>
 #include <unistd.h>
+#include <sys/wait.h>
 
 struct _TerminalScreenPrivate
 {
@@ -56,6 +58,7 @@ struct _TerminalScreenPrivate
   GtkWidget *title_entry;
   char *working_dir;
   int child_pid;
+  double font_scale;
   guint recheck_working_dir_idle;
 };
 
@@ -65,6 +68,7 @@ enum {
   PROFILE_SET,
   TITLE_CHANGED,
   SELECTION_CHANGED,
+  ENCODING_CHANGED,
   LAST_SIGNAL
 };
 
@@ -73,14 +77,12 @@ static void terminal_screen_class_init  (TerminalScreenClass *klass);
 static void terminal_screen_finalize    (GObject             *object);
 static void terminal_screen_update_on_realize (GtkWidget      *widget,
                                                TerminalScreen *screen);
-
 static void     terminal_screen_popup_menu         (GtkWidget      *term,
                                                     TerminalScreen *screen);
 static gboolean terminal_screen_button_press_event (GtkWidget      *term,
                                                     GdkEventButton *event,
                                                     TerminalScreen *screen);
 
-
 static void terminal_screen_widget_title_changed     (GtkWidget      *term,
                                                       TerminalScreen *screen);
 
@@ -90,6 +92,9 @@ static void terminal_screen_widget_child_died        (GtkWidget      *term,
 static void terminal_screen_widget_selection_changed (GtkWidget      *term,
                                                       TerminalScreen *screen);
 
+static void terminal_screen_widget_encoding_changed  (GtkWidget      *term,
+                                                      TerminalScreen *screen);
+
 static void terminal_screen_setup_dnd                (TerminalScreen *screen);
 
 static void reread_profile (TerminalScreen *screen);
@@ -140,6 +145,16 @@ terminal_screen_get_type (void)
   return object_type;
 }
 
+
+static void
+style_set_callback (GtkWidget *widget,
+                    GtkStyle  *previous_style,
+                    void      *data)
+{
+  if (GTK_WIDGET_REALIZED (widget))
+    terminal_screen_update_on_realize (widget, TERMINAL_SCREEN (data));  
+}
+
 static void
 terminal_screen_init (TerminalScreen *screen)
 {
@@ -155,6 +170,8 @@ terminal_screen_init (TerminalScreen *screen)
   screen->priv->recheck_working_dir_idle = 0;
   
   screen->priv->term = terminal_widget_new ();
+
+  screen->priv->font_scale = PANGO_SCALE_MEDIUM;
   
   g_object_ref (G_OBJECT (screen->priv->term));
   gtk_object_sink (GTK_OBJECT (screen->priv->term));
@@ -176,6 +193,11 @@ terminal_screen_init (TerminalScreen *screen)
                     G_CALLBACK (terminal_screen_update_on_realize),
                     screen);
 
+  g_signal_connect (G_OBJECT (screen->priv->term),
+                    "style_set",
+                    G_CALLBACK (style_set_callback),
+                    screen);
+  
   g_signal_connect (G_OBJECT (screen->priv->term),
                     "popup_menu",
                     G_CALLBACK (terminal_screen_popup_menu),
@@ -198,6 +220,10 @@ terminal_screen_init (TerminalScreen *screen)
                                              G_CALLBACK (terminal_screen_widget_selection_changed),
                                              screen);
 
+  terminal_widget_connect_encoding_changed (screen->priv->term,
+                                            G_CALLBACK (terminal_screen_widget_encoding_changed),
+                                            screen);
+
   gtk_widget_show (screen->priv->term);
 }
 
@@ -209,7 +235,7 @@ terminal_screen_class_init (TerminalScreenClass *klass)
   parent_class = g_type_class_peek_parent (klass);
   
   object_class->finalize = terminal_screen_finalize;
-
+  
   signals[PROFILE_SET] =
     g_signal_new ("profile_set",
                   G_OBJECT_CLASS_TYPE (object_class),
@@ -236,6 +262,15 @@ terminal_screen_class_init (TerminalScreenClass *klass)
                   NULL, NULL,
                   g_cclosure_marshal_VOID__VOID,
                   G_TYPE_NONE, 0);  
+
+  signals[ENCODING_CHANGED] =
+    g_signal_new ("encoding_changed",
+                  G_OBJECT_CLASS_TYPE (object_class),
+                  G_SIGNAL_RUN_LAST,
+                  G_STRUCT_OFFSET (TerminalScreenClass, encoding_changed),
+                  NULL, NULL,
+                  g_cclosure_marshal_VOID__VOID,
+                  G_TYPE_NONE, 0);  
 }
 
 static void
@@ -372,6 +407,15 @@ reread_profile (TerminalScreen *screen)
   terminal_widget_set_scrollback_lines (term,
                                         terminal_profile_get_scrollback_lines (profile));
 
+  if (terminal_profile_get_use_skey (screen->priv->profile))
+    {
+      terminal_widget_skey_match_add (screen->priv->term,
+				      "s/key ([0-9]*) ([-A-Za-z0-9]*)");
+    }
+  else
+    {
+      terminal_widget_skey_match_remove (screen->priv->term);
+    }
   bg_type = terminal_profile_get_background_type (profile);
   
   if (bg_type == TERMINAL_BACKGROUND_IMAGE)
@@ -453,9 +497,9 @@ rebuild_title  (TerminalScreen *screen)
 }
 
 static void
-profile_changed_callback (TerminalProfile          *profile,
-                          TerminalSettingMask       mask,
-                          TerminalScreen           *screen)
+profile_changed_callback (TerminalProfile           *profile,
+                          const TerminalSettingMask *mask,
+                          TerminalScreen            *screen)
 {
   reread_profile (screen);
 }
@@ -488,6 +532,9 @@ update_color_scheme (TerminalScreen *screen)
                               &fg, &bg, palette);
 }
 
+/* Note this can be called on style_set, on prefs changing,
+ * and on setting the font scale factor
+ */
 static void
 terminal_screen_update_on_realize (GtkWidget      *term,
                                    TerminalScreen *screen)
@@ -525,20 +572,33 @@ terminal_screen_update_on_realize (GtkWidget      *term,
                                                    pango_font_description_get_size (term->style->font_desc));
                 }
 
-              terminal_widget_set_pango_font (term,
-                                              terminal_profile_get_font (profile));
+              pango_font_description_set_size (desc,
+                                               screen->priv->font_scale *
+                                               pango_font_description_get_size (desc));
+              
+              terminal_widget_set_pango_font (term, desc);
 
               pango_font_description_free (desc);
             }
         }
       else
         {
-          terminal_widget_set_pango_font (term,
-                                          terminal_profile_get_font (profile));
+          PangoFontDescription *desc;
+
+          desc = pango_font_description_copy (terminal_profile_get_font (profile));
+          pango_font_description_set_size (desc,
+                                           screen->priv->font_scale *
+                                           pango_font_description_get_size (desc));          
+          
+          terminal_widget_set_pango_font (term, desc);
+
+          pango_font_description_free (desc);
         }
     }
   else
     {
+      /* font_scale is not supported in X fonts mode */
+      
       GdkFont *font;
 
       font = NULL;
@@ -625,7 +685,7 @@ profile_forgotten_callback (TerminalProfile *profile,
   TerminalProfile *new_profile;
 
   /* Revert to the new term profile if any */
-  new_profile = terminal_profile_get_for_new_term ();
+  new_profile = terminal_profile_get_for_new_term (NULL);
 
   if (new_profile)
     terminal_screen_set_profile (screen, new_profile);
@@ -805,7 +865,7 @@ get_child_command (TerminalScreen *screen,
   else
     {
       const char *only_name;
-      const char *shell;
+      char *shell;
 
       shell = gnome_util_user_shell ();
 
@@ -825,6 +885,8 @@ get_child_command (TerminalScreen *screen,
         argv[0] = g_strdup (only_name);
 
       argv[1] = NULL;
+
+      g_free (shell);
     }
 
   if (file_p)
@@ -851,7 +913,7 @@ get_child_environment (GtkWidget      *term,
   char **p;
   int i;
   char **retval;
-#define EXTRA_ENV_VARS 6
+#define EXTRA_ENV_VARS 7
   
   profile = screen->priv->profile;
 
@@ -870,7 +932,8 @@ get_child_environment (GtkWidget      *term,
           (strncmp (*p, "WINDOWID=", 9) == 0) ||
           (strncmp (*p, "TERM=", 5) == 0)    ||
           (strncmp (*p, "GNOME_DESKTOP_ICON=", 19) == 0) ||
-          (strncmp (*p, "COLORTERM=", 10) == 0))
+          (strncmp (*p, "COLORTERM=", 10) == 0) ||
+	  (strncmp (*p, "DISPLAY=", 8) == 0))
         {
           /* nothing: do not copy */
         }
@@ -888,6 +951,9 @@ get_child_environment (GtkWidget      *term,
   retval[i] = g_strdup_printf ("WINDOWID=%ld",
                                GDK_WINDOW_XWINDOW (term->window));
   ++i;
+  retval[i] = g_strdup_printf ("DISPLAY=%s", 
+				gdk_display_get_name(gtk_widget_get_display(term)));
+  ++i;
   
   retval[i] = NULL;
   
@@ -956,10 +1022,17 @@ static void
 new_window_callback (GtkWidget      *menu_item,
                      TerminalScreen *screen)
 {
+  char *name;
+
+  name = gdk_screen_make_display_name (gtk_widget_get_screen (menu_item));
+  
   terminal_app_new_terminal (terminal_app_get (),
-                             terminal_profile_get_for_new_term (),
+                             terminal_profile_get_for_new_term (screen->priv->profile),
                              NULL,
-                             FALSE, FALSE, NULL, NULL, NULL, NULL);
+                             FALSE, FALSE, NULL, NULL, NULL, NULL, NULL, 1.0,
+                             NULL, name, -1);
+
+  g_free (name);
 }
 
 static void
@@ -967,9 +1040,10 @@ new_tab_callback (GtkWidget      *menu_item,
                   TerminalScreen *screen)
 {
   terminal_app_new_terminal (terminal_app_get (),
-                             terminal_profile_get_for_new_term (),
+                             terminal_profile_get_for_new_term (screen->priv->profile),
                              screen->priv->window,
-                             FALSE, FALSE, NULL, NULL, NULL, NULL);
+                             FALSE, FALSE, NULL, NULL, NULL, NULL, NULL, 1.0,
+                             NULL, NULL, -1);
 }
 
 static void
@@ -1171,6 +1245,7 @@ terminal_screen_do_popup (TerminalScreen *screen,
                           GdkEventButton *event)
 {
   GtkWidget *profile_menu;
+  GtkWidget *im_menu;
   GtkWidget *menu_item;
   GList *profiles;
   GList *tmp;
@@ -1276,6 +1351,21 @@ terminal_screen_do_popup (TerminalScreen *screen,
 		   screen);
 
   
+  im_menu = gtk_menu_new ();
+  menu_item = gtk_menu_item_new_with_mnemonic (_("_Input Methods"));
+
+  gtk_menu_item_set_submenu (GTK_MENU_ITEM (menu_item),
+                             im_menu);
+
+  terminal_widget_im_append_menuitems (screen->priv->term,
+		  		       GTK_MENU_SHELL (im_menu));
+
+  gtk_widget_show (im_menu);
+  gtk_widget_show (menu_item);
+
+  gtk_menu_shell_append (GTK_MENU_SHELL (screen->priv->popup_menu),
+                         menu_item);
+
   if (terminal_window_get_menubar_visible (screen->priv->window))
     menu_item = append_menuitem (screen->priv->popup_menu,
 				 _("Hide _Menubar"),
@@ -1332,6 +1422,7 @@ terminal_screen_button_press_event (GtkWidget      *widget,
 {
   GtkWidget *term;
   int char_width, char_height;
+  gboolean dingus_button;
   
   term = screen->priv->term;
 
@@ -1342,43 +1433,47 @@ terminal_screen_button_press_event (GtkWidget      *widget,
     terminal_widget_check_match (term,
                                  event->x / char_width,
                                  event->y / char_height);
-  
-  if (event->button == 1 || event->button == 2)
+  dingus_button = ((event->button == 1) || (event->button == 2));
+
+  if (dingus_button &&
+      (event->state & GDK_CONTROL_MASK) &&
+      terminal_profile_get_use_skey (screen->priv->profile))
+    {
+      gchar *skey_match;
+
+      skey_match = terminal_widget_skey_check_match (term,
+						     event->x / char_width,
+						     event->y / char_height);
+      if (skey_match != NULL)
+	{
+	  terminal_skey_do_popup (screen, GTK_WINDOW (terminal_screen_get_window (screen)), skey_match);
+	  g_free (skey_match);
+
+	  return TRUE;
+	}
+    }
+
+  if (dingus_button &&
+      (event->state & GDK_CONTROL_MASK) &&
+      (screen->priv->matched_string != NULL))
     {
       gtk_widget_grab_focus (widget);
       
-      if (screen->priv->matched_string &&
-          (event->state & GDK_CONTROL_MASK))
-        {
-          open_url (screen, screen->priv->matched_string);
-          g_free (screen->priv->matched_string);
-          screen->priv->matched_string = NULL;
-          return TRUE; /* don't do anything else such as select with the click */
-        }
-      
-#if 0
-      /* old gnome-terminal had this code, but I'm not sure if
-       * it should be here. We always return FALSE, but
-       * sometimes old gnome-terminal didn't, so maybe that's
-       * why it had this.
-       */
-      if (event->button != 3
-          || (!(event->state & GDK_CONTROL_MASK) && term->vx->selected)
-          || (term->vx->vt.mode & VTMODE_SEND_MOUSE))
-        return FALSE;
-#endif      
-
-      return FALSE; /* pass thru the click */
+      open_url (screen, screen->priv->matched_string);
+      g_free (screen->priv->matched_string);
+      screen->priv->matched_string = NULL;
+      return TRUE; /* don't do anything else such as select with the click */
     }
-  else if (event->button == 3)
+      
+  if ((event->button == 3) &&
+      !(event->state & (GDK_SHIFT_MASK | GDK_CONTROL_MASK | GDK_MOD1_MASK)))
     {
       terminal_screen_do_popup (screen, event);
       return TRUE;
     }
-  else
-    {
-      return FALSE;
-    }
+
+  /* default behavior is to let the terminal widget deal with it */
+  return FALSE;
 }
 
 void
@@ -1434,27 +1529,80 @@ terminal_screen_get_working_dir (TerminalScreen *screen)
 {
   g_return_val_if_fail (TERMINAL_IS_SCREEN (screen), NULL);
 
-  /* If we're on Linux and have a child PID, try to update
-   * the working dir
-   */
+  /* Try to update the working dir using various OS-specific mechanisms */
   if (screen->priv->child_pid >= 0)
     {
       char *file;
       char buf[PATH_MAX+1];
       int len;
-      
+
+      /* readlink (/proc/pid/cwd) will work on Linux */
       file = g_strdup_printf ("/proc/%d/cwd", screen->priv->child_pid);
 
       /* Silently ignore failure here, since we may not be on Linux */
       len = readlink (file, buf, sizeof (buf) - 1);
 
-      if (len >= 0)
+      /*
+       * If readlink fails, get the current working directory
+       * using the command 'pwdx'. This is specific to Solaris.
+       */
+      if (len > 0 && buf[0] == '/')
         {
           buf[len] = '\0';
           
           g_free (screen->priv->working_dir);
           screen->priv->working_dir = g_strdup (buf);
         }
+      else
+        {
+          char *abs_command = NULL;
+          char *exec_command = NULL;
+          char *output = NULL;
+          int exit_status = 0;
+          
+          abs_command = g_find_program_in_path ("pwdx");
+          if (abs_command == NULL)
+            goto pwdx_out;
+
+          exec_command = g_strdup_printf ("%s %d", abs_command, 
+                                          screen->priv->child_pid);
+               
+          if (!g_spawn_command_line_sync (exec_command, &output, 
+                                          NULL, &exit_status, NULL))
+            goto pwdx_out;          
+               
+          if (WIFEXITED (exit_status) &&
+              WEXITSTATUS (exit_status) == 0 &&
+              output) 
+            {
+              /*
+               * 'pwdx' output format is:
+               * <pid>:\t<cwd>\n
+               */
+              char *p;
+
+              p = strstr (output, ":");
+              if (p == NULL)
+                goto pwdx_out;
+
+              ++p;
+                   
+              while (*p == ' ' || *p == '\t')
+                ++p;
+              g_strchomp (p);
+                   
+              if (*p == '/')
+                {
+                  g_free (screen->priv->working_dir);
+                  screen->priv->working_dir = g_strdup (p);
+                }
+            }
+
+        pwdx_out:
+          g_free (abs_command);
+          g_free (exec_command);
+          g_free (output);
+        }
       
       g_free (file);
     }
@@ -1489,6 +1637,37 @@ queue_recheck_working_dir (TerminalScreen *screen)
     }
 }
 
+
+void
+terminal_screen_set_font_scale (TerminalScreen *screen,
+                                double          factor)
+{
+  g_return_if_fail (TERMINAL_IS_SCREEN (screen));
+
+  if (factor < TERMINAL_SCALE_MINIMUM)
+    factor = TERMINAL_SCALE_MINIMUM;
+  if (factor > TERMINAL_SCALE_MAXIMUM)
+    factor = TERMINAL_SCALE_MAXIMUM;
+  
+  screen->priv->font_scale = factor;
+  
+  if (screen->priv->term &&
+      GTK_WIDGET_REALIZED (screen->priv->term))
+    {
+      /* Update the font */
+      terminal_screen_update_on_realize (screen->priv->term,
+                                         screen);
+    }
+}
+
+double
+terminal_screen_get_font_scale (TerminalScreen *screen)
+{
+  g_return_val_if_fail (TERMINAL_IS_SCREEN (screen), 1.0);
+  
+  return screen->priv->font_scale;
+}
+
 static void
 terminal_screen_widget_title_changed (GtkWidget      *widget,
                                       TerminalScreen *screen)
@@ -1529,6 +1708,13 @@ terminal_screen_widget_selection_changed (GtkWidget      *term,
   g_signal_emit (G_OBJECT (screen), signals[SELECTION_CHANGED], 0);
 }
 
+static void
+terminal_screen_widget_encoding_changed (GtkWidget      *term,
+                                         TerminalScreen *screen)
+{
+  g_signal_emit (G_OBJECT (screen), signals[ENCODING_CHANGED], 0);
+}
+
 static void
 title_entry_changed (GtkWidget      *entry,
                      TerminalScreen *screen)
@@ -1686,15 +1872,14 @@ drag_data_received  (GtkWidget *widget, GdkDragContext *context,
         
         str = gtk_selection_data_get_text (selection_data);
 
-        /* FIXME should this be converted back to locale encoding
-         * before sending to the terminal? Most likely yes.
+        /*
+	 * pass UTF-8 to the terminal widget. The terminal widget
+	 * should know which encoding mode it's in and be able
+	 * to perform the correct conversion.
          */
         if (str && *str)
-          {
-            terminal_widget_write_data_to_child (screen->priv->term,
-                                                 str, strlen (str));
-          }
-
+	  terminal_widget_write_data_to_child (screen->priv->term,
+					       str, strlen (str));
         g_free (str);
       }
       break;
@@ -1969,14 +2154,14 @@ terminal_screen_setup_dnd (TerminalScreen *screen)
                      GDK_ACTION_COPY);
 }
 
-/* Indices of some xlfd string entries.  The first entry is 0 */
-#define XLFD_WEIGHT_INDEX			2
-#define XLFD_SLANT_INDEX			3
-#define XLFD_SIZE_IN_PIXELS_INDEX		6
-#define XLFD_SIZE_IN_POINTS_INDEX		7
-#define XLFD_HORIZONTAL_RESOLUTION_INDEX	8
-#define XLFD_VERTICAL_RESOLUTION_INDEX		9
-#define XLFD_SPACING_INDEX		        10
+/* Indices of some xlfd string entries.  The first entry is 1 */
+#define XLFD_WEIGHT_INDEX			3
+#define XLFD_SLANT_INDEX			4	
+#define XLFD_SIZE_IN_PIXELS_INDEX		7	
+#define XLFD_SIZE_IN_POINTS_INDEX		8	
+#define XLFD_HORIZONTAL_RESOLUTION_INDEX	9	
+#define XLFD_VERTICAL_RESOLUTION_INDEX		10	
+#define XLFD_SPACING_INDEX		        11
 
 #define XLFD_N_FIELDS 14
 
@@ -2096,7 +2281,7 @@ make_xfont_have_size_from_other_font (const char *fontname,
 
   ret = NULL;
   size_pixels = xlfd_get_nth_field (other_font, XLFD_SIZE_IN_PIXELS_INDEX);
-  size_points = xlfd_get_nth_field (other_font, XLFD_SIZE_IN_PIXELS_INDEX);
+  size_points = xlfd_get_nth_field (other_font, XLFD_SIZE_IN_POINTS_INDEX);
 
   if (size_pixels && size_points)
     {
@@ -2128,13 +2313,10 @@ make_font_monospace (const PangoFontDescription *font)
 {
   PangoFontDescription *ret;
 
-  return NULL;
-  
   ret = pango_font_description_copy (font);
 
-  /* FIXME use a heuristic here such as getting width of "i" and
-   * "M" and if they match our font is monospace.
-   */
+  /* FIXME */
+  pango_font_description_set_family (ret, "monospace");
   
   return ret;
 }
